// Copyright (c) 2013 by István Váradi

// This file is part of XPLRA, a remote-access plugin for X-Plane

// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:

// 1. Redistributions of source code must retain the above copyright notice, this
//    list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// The views and conclusions contained in the software and documentation are those
// of the authors and should not be interpreted as representing official policies,
// either expressed or implied, of the FreeBSD Project.

//------------------------------------------------------------------------------

#include "MultiGetter.h"

#include "XPlane.h"

#include <xplra/Protocol.h>

#include <hu/varadiistvan/scpl/io/DataStream.h>

#include <cstdio>

//------------------------------------------------------------------------------

using hu::varadiistvan::xplra::MultiGetter;

using xplra::Protocol;

using std::vector;
using std::min;

//------------------------------------------------------------------------------

MultiGetter::MultiGetter(XPlane& xplane) throw() :
    MultiBuffer(xplane, Protocol::COMMAND_REGISTER_GET_MULTI,
                Protocol::COMMAND_UNREGISTER_GET_MULTI)
{
}

//------------------------------------------------------------------------------

inline void MultiGetter::read(const DataRef& dataRef) throw(Exception)
{
    switch (dataRef.type) {
      case Protocol::TYPE_INT:
        *reinterpret_cast<int32_t*>(data + dataRef.dataOffset) =
            xplane.stream->readS32();
        break;
      case Protocol::TYPE_FLOAT:
        *reinterpret_cast<float*>(data + dataRef.dataOffset) =
            xplane.stream->readFloat();
        break;
      case Protocol::TYPE_DOUBLE:
        *reinterpret_cast<double*>(data + dataRef.dataOffset) =
            xplane.stream->readDouble();
        break;
      case Protocol::TYPE_FLOAT_ARRAY:
      case Protocol::TYPE_INT_ARRAY:
      case Protocol::TYPE_BYTE_ARRAY:
      {
          size_t length = xplane.stream->readS32();
          xplane.checkStream();
          length = min(length, dataRef.length);
          size_t elementSize =
              (dataRef.type==Protocol::TYPE_FLOAT_ARRAY) ? sizeof(float) :
              ((dataRef.type==Protocol::TYPE_INT_ARRAY) ?
               sizeof(int32_t) : sizeof(uint8_t));
          xplane.stream->read(data + dataRef.dataOffset, length * elementSize);
      }
    }
}

//------------------------------------------------------------------------------

void MultiGetter::doExecute() throw(Exception)
{
    xplane.checkStream();
    xplane.stream->writeU8(Protocol::COMMAND_EXECUTE_GET_MULTI);
    xplane.stream->writeU32(registeredID);
    xplane.stream->flush();
    xplane.checkResult();

    readValues();
}


//------------------------------------------------------------------------------

void MultiGetter::doExecuteUnregistered() throw(Exception)
{
    writeSpec(Protocol::COMMAND_GET_MULTI);

    readValues();
}

//------------------------------------------------------------------------------

void MultiGetter::readValues() throw(Exception)
{
    for(vector<DataRef>::const_iterator i = dataRefs.begin();
        i!=dataRefs.end(); ++i)
    {
        xplane.checkStream();
        const DataRef& dataRef = *i;
        read(dataRef);
    }
}

//------------------------------------------------------------------------------

// Local Variables:
// mode: C++
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
